# -*- encoding: utf-8 -*-

"""
A List of Useful Functions to Provide Functionality over a Dataframe

The :attr:`numpy` and :attr:`pandas` has various in-built functions
to quickly summarize dataframe. The functions tabs enables end-users
with the capability of creating custom functions (like half-life) and
apply the same over a pandas dataframe object.
"""

from pandaswizard.functions.collate import * # noqa: F401, F403 # pyright: ignore[reportMissingImports]
